# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import io

from drgn import Program, TypeEnumerator
from drgn.helpers.common.format import (
    CellFormat,
    RowOptions,
    decode_enum_type_flags,
    decode_flags,
    number_in_binary_units,
    print_table,
)
from tests import MOCK_PLATFORM, TestCase


class TestDecodeFlags(TestCase):
    FLAGS_BIT_NUMBERS = (("BOLD", 0), ("ITALIC", 1), ("UNDERLINE", 2))
    FLAGS_VALUES = tuple(
        (name, 1 << bit_number) for name, bit_number in FLAGS_BIT_NUMBERS
    )

    def assertDecodeFlags(self, value, expected):
        self.assertEqual(decode_flags(value, self.FLAGS_BIT_NUMBERS), expected)
        self.assertEqual(decode_flags(value, self.FLAGS_VALUES, False), expected)

    def test_one(self):
        self.assertDecodeFlags(1, "BOLD")
        self.assertDecodeFlags(2, "ITALIC")
        self.assertDecodeFlags(4, "UNDERLINE")

    def test_multiple(self):
        self.assertDecodeFlags(3, "BOLD|ITALIC")
        self.assertDecodeFlags(5, "BOLD|UNDERLINE")
        self.assertDecodeFlags(6, "ITALIC|UNDERLINE")
        self.assertDecodeFlags(7, "BOLD|ITALIC|UNDERLINE")

    def test_all_unknown(self):
        self.assertDecodeFlags(8, "0x8")

    def test_some_unknown(self):
        self.assertDecodeFlags(9, "BOLD|0x8")

    def test_alias(self):
        self.assertEqual(
            decode_flags(2, (("SMALL", 0), ("BIG", 1), ("LARGE", 1))), "BIG|LARGE"
        )
        self.assertEqual(
            decode_flags(2, (("SMALL", 1), ("BIG", 2), ("LARGE", 2)), False),
            "BIG|LARGE",
        )

    def test_no_aliases(self):
        self.assertEqual(
            decode_flags(2, (("SMALL", 0), ("BIG", 1), ("LARGE", 1)), aliases=False),
            "BIG",
        )

    def test_no_aliases_all_unknown(self):
        self.assertEqual(
            decode_flags(8, (("SMALL", 0), ("BIG", 1), ("LARGE", 1)), aliases=False),
            "0x8",
        )

    def test_no_aliases_some_unknown(self):
        self.assertEqual(
            decode_flags(10, (("SMALL", 0), ("BIG", 1), ("LARGE", 1)), aliases=False),
            "BIG|0x8",
        )

    def test_zero(self):
        self.assertDecodeFlags(0, "0")

    def test_decode_enum_type_flags(self):
        prog = Program(MOCK_PLATFORM)

        for bit_numbers, flags in (
            (True, self.FLAGS_BIT_NUMBERS),
            (False, self.FLAGS_VALUES),
        ):
            with self.subTest(bit_numbers=bit_numbers):
                type = prog.enum_type(
                    None,
                    prog.int_type("int", 4, True),
                    [TypeEnumerator(*flag) for flag in flags],
                )
                self.assertEqual(
                    decode_enum_type_flags(4, type, bit_numbers), "UNDERLINE"
                )
                self.assertEqual(
                    decode_enum_type_flags(27, type, bit_numbers), "BOLD|ITALIC|0x18"
                )

    def test_decode_enum_type_flags_incomplete(self):
        self.assertRaisesRegex(
            TypeError,
            "incomplete",
            decode_enum_type_flags,
            2,
            Program().enum_type(None),
        )


class TestNumberInBinaryUnits(TestCase):
    def test_zero(self):
        self.assertEqual(number_in_binary_units(0), "0")

    def test_small(self):
        self.assertEqual(number_in_binary_units(100), "100")
        self.assertEqual(number_in_binary_units(1023), "1023")

    def test_small_negative(self):
        self.assertEqual(number_in_binary_units(-100), "-100")
        self.assertEqual(number_in_binary_units(-1023), "-1023")

    def test_integer(self):
        self.assertEqual(number_in_binary_units(1024), "1K")
        self.assertEqual(number_in_binary_units(1024**2), "1M")
        self.assertEqual(number_in_binary_units(1024**3), "1G")
        self.assertEqual(number_in_binary_units(1024**4), "1T")
        self.assertEqual(number_in_binary_units(1024**5), "1P")
        self.assertEqual(number_in_binary_units(1024**6), "1E")
        self.assertEqual(number_in_binary_units(1024**7), "1Z")
        self.assertEqual(number_in_binary_units(1024**8), "1Y")

    def test_negative_integer(self):
        self.assertEqual(number_in_binary_units(-1024), "-1K")
        self.assertEqual(number_in_binary_units(-(1024**2)), "-1M")
        self.assertEqual(number_in_binary_units(-(1024**3)), "-1G")
        self.assertEqual(number_in_binary_units(-(1024**4)), "-1T")
        self.assertEqual(number_in_binary_units(-(1024**5)), "-1P")
        self.assertEqual(number_in_binary_units(-(1024**6)), "-1E")
        self.assertEqual(number_in_binary_units(-(1024**7)), "-1Z")
        self.assertEqual(number_in_binary_units(-(1024**8)), "-1Y")

    def test_almost_integer(self):
        self.assertEqual(number_in_binary_units(1025), "1.0K")
        self.assertEqual(number_in_binary_units(1024**4 + 1), "1.0T")

    def test_precision(self):
        n = 1088
        self.assertEqual(number_in_binary_units(n, precision=0), "1K")
        self.assertEqual(number_in_binary_units(n, precision=1), "1.1K")
        self.assertEqual(number_in_binary_units(n, precision=2), "1.06K")

    def test_huge(self):
        self.assertEqual(number_in_binary_units(1024**8 * 1.5), "1.5Y")
        self.assertEqual(number_in_binary_units(1024**10), "1048576Y")


class TestPrintTable(TestCase):
    def assert_print_table(self, rows, expected, **kwargs):
        f = io.StringIO()
        print_table(rows, file=f)
        self.assertEqual(f.getvalue(), expected)

    def test_empty(self):
        self.assert_print_table([], "")

    def test_one_row(self):
        self.assert_print_table([["abc", "de", "fghi"]], "abc  de  fghi\n")

    def test_align(self):
        self.assert_print_table(
            [[2, 2000, 4], [13, 3, 19]],
            """\
 2  2000   4
13     3  19
""",
        )

    def test_empty_cell(self):
        self.assert_print_table(
            [[2, 2000, 4], ["", 3, 13, 19]],
            """\
2  2000   4
      3  13  19
""",
        )

    def test_cell_format(self):
        self.assert_print_table(
            [
                ["DECIMAL", "HEXADECIMAL"],
                [CellFormat(10, "<"), CellFormat(10, "<x")],
            ],
            """\
DECIMAL  HEXADECIMAL
10       a
""",
        )

    def test_row_options(self):
        self.assert_print_table(
            [
                ["LETTER", "INDEX"],
                ["A", 1],
                RowOptions(["WORD", "LENGTH"], group=1),
                RowOptions(["angstrom", 8], group=1),
                ["LETTER", "INDEX"],
                ["B", 2],
                RowOptions(["WORD", "LENGTH"], group=1),
                RowOptions(["banana", 6], group=1),
            ],
            """\
LETTER  INDEX
A           1
WORD      LENGTH
angstrom       8
LETTER  INDEX
B           2
WORD      LENGTH
banana         6
""",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import Object
from drgn.helpers.linux.kthread import kthread_data, task_is_kthread
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestKthread(LinuxKernelTestCase):
    # There's no good way to test to_kthread() directly, but it gets tested
    # indirectly through kthread_data().
    def test_kthread_data(self):
        self.assertIdentical(
            kthread_data(self.prog["drgn_test_kthread"]),
            Object(self.prog, "void *", 0xB0BA000),
        )

    def test_task_is_kthread(self):
        self.assertTrue(task_is_kthread(self.prog["init_task"].address_of_()))
        self.assertFalse(task_is_kthread(find_task(self.prog, 1)))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from pathlib import Path

import drgn.helpers.linux.cpumask
from drgn.helpers.linux.cpumask import cpumask_of, cpumask_to_cpulist, for_each_cpu
from tests.linux_kernel import LinuxKernelTestCase, parse_range_list, possible_cpus

CPU_PATH = Path("/sys/devices/system/cpu")


class TestCpuMask(LinuxKernelTestCase):
    _MASKS = ("online", "possible", "present")

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        for online_path in sorted(CPU_PATH.glob("cpu*/online")):
            if int(online_path.read_text()):
                cls.offlined_path = online_path
                online_path.write_text("0")
                cls.addClassCleanup(online_path.write_text, "1")
                break

    def test_for_each_cpu(self):
        for name in self._MASKS:
            with self.subTest(name=name):
                self.assertEqual(
                    list(
                        getattr(drgn.helpers.linux.cpumask, f"for_each_{name}_cpu")(
                            self.prog
                        )
                    ),
                    sorted(parse_range_list((CPU_PATH / name).read_text())),
                )

    def test_cpumask_to_cpulist(self):
        for name in self._MASKS:
            with self.subTest(name=name):
                self.assertEqual(
                    cpumask_to_cpulist(
                        getattr(drgn.helpers.linux.cpumask, f"cpu_{name}_mask")(
                            self.prog
                        )
                    ),
                    (CPU_PATH / name).read_text().strip(),
                )

    def test_cpumask_weight(self):
        for name in self._MASKS:
            with self.subTest(name=name):
                self.assertEqual(
                    getattr(drgn.helpers.linux.cpumask, f"num_{name}_cpus")(self.prog),
                    len(parse_range_list((CPU_PATH / name).read_text())),
                )

    def test_cpumask_of(self):
        cpu = max(possible_cpus())
        self.assertEqual(list(for_each_cpu(cpumask_of(self.prog, cpu))), [cpu])



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from pathlib import Path
import re

from drgn.helpers.linux.ioport import for_each_resource
from tests.linux_kernel import LinuxKernelTestCase


class TestIoport(LinuxKernelTestCase):
    def test_for_each_resource(self):
        resources = for_each_resource(self.prog["ioport_resource"].address_of_())
        next(resources)  # Skip the root.
        self.assertEqual(
            [resource.name.string_() for resource in resources],
            re.findall(b"[^:]* : (.*)", Path("/proc/ioports").read_bytes()),
        )



# Copyright (c) 2022, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import NULL
from drgn.helpers.linux.llist import (
    llist_empty,
    llist_first_entry,
    llist_first_entry_or_null,
    llist_for_each,
    llist_for_each_entry,
    llist_is_singular,
    llist_next_entry,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestLlist(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.empty = cls.prog["drgn_test_empty_llist"].address_of_()
        cls.full = cls.prog["drgn_test_full_llist"].address_of_()
        cls.entries = cls.prog["drgn_test_llist_entries"]
        cls.num_entries = 3
        cls.singular = cls.prog["drgn_test_singular_llist"].address_of_()
        cls.singular_entry = cls.prog["drgn_test_singular_llist_entry"].address_of_()
        cls.singular_node = cls.singular_entry.node.address_of_()

    def node(self, n):
        return self.entries[n].node.address_of_()

    def entry(self, n):
        return self.entries[n].address_of_()

    def test_llist_empty(self):
        self.assertTrue(llist_empty(self.empty))
        self.assertFalse(llist_empty(self.full))
        self.assertFalse(llist_empty(self.singular))

    def test_llist_is_singular(self):
        self.assertFalse(llist_is_singular(self.empty))
        self.assertFalse(llist_is_singular(self.full))
        self.assertTrue(llist_is_singular(self.singular))

    def test_llist_first_entry(self):
        self.assertEqual(
            llist_first_entry(self.full, "struct drgn_test_llist_entry", "node"),
            self.entry(2),
        )
        self.assertEqual(
            llist_first_entry(self.singular, "struct drgn_test_llist_entry", "node"),
            self.singular_entry,
        )

    def test_llist_first_entry_or_null(self):
        self.assertEqual(
            llist_first_entry_or_null(
                self.empty, "struct drgn_test_llist_entry", "node"
            ),
            NULL(self.prog, "struct drgn_test_llist_entry *"),
        )
        self.assertEqual(
            llist_first_entry_or_null(
                self.full, "struct drgn_test_llist_entry", "node"
            ),
            self.entry(2),
        )
        self.assertEqual(
            llist_first_entry_or_null(
                self.singular, "struct drgn_test_llist_entry", "node"
            ),
            self.singular_entry,
        )

    def test_llist_next_entry(self):
        for i in reversed(range(1, self.num_entries)):
            self.assertEqual(llist_next_entry(self.entry(i), "node"), self.entry(i - 1))

    def test_llist_for_each(self):
        self.assertEqual(list(llist_for_each(self.empty.first)), [])
        self.assertEqual(
            list(llist_for_each(self.full.first)),
            [self.node(i) for i in reversed(range(self.num_entries))],
        )
        self.assertEqual(
            list(llist_for_each(self.singular.first)), [self.singular_node]
        )

    def test_llist_for_each_entry(self):
        self.assertEqual(
            list(
                llist_for_each_entry(
                    "struct drgn_test_llist_entry", self.empty.first, "node"
                )
            ),
            [],
        )
        self.assertEqual(
            list(
                llist_for_each_entry(
                    "struct drgn_test_llist_entry", self.full.first, "node"
                )
            ),
            [self.entry(i) for i in reversed(range(self.num_entries))],
        )
        self.assertEqual(
            list(
                llist_for_each_entry(
                    "struct drgn_test_llist_entry", self.singular.first, "node"
                )
            ),
            [self.singular_entry],
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

REGISTERS = [
    *[DrgnRegister(f"r{i}") for i in range(16)],
    *[DrgnRegister(f"a{i}") for i in range(16)],
    DrgnRegister("pswm"),
    DrgnRegister("pswa"),
]

REGISTER_LAYOUT = [
    # Callee-saved registers and return address (r14).
    *[DrgnRegisterLayout(f"r{i}", size=8, dwarf_number=i) for i in range(6, 16)],
    # Caller-saved registers.
    *[DrgnRegisterLayout(f"r{i}", size=8, dwarf_number=i) for i in range(6)],
    # These are typically only used for interrupted frames.
    DrgnRegisterLayout("pswm", size=8, dwarf_number=64),
    DrgnRegisterLayout("pswa", size=8, dwarf_number=65),
    # Access control registers (ACRs) are only used in userspace,
    # and not present in struct pt_regs.
    *[DrgnRegisterLayout(f"a{i}", size=4, dwarf_number=48 + i) for i in range(16)],
]

STACK_POINTER_REGISTER = "r15"



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from pathlib import Path
import unittest

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from drgn.helpers.linux.cpumask import for_each_cpu
from drgn.helpers.linux.irq import (
    for_each_irq_desc,
    gate_desc_func,
    irq_desc_action_names,
    irq_desc_affinity_mask,
    irq_desc_chip_name,
    irq_desc_kstat_cpu,
    irq_to_desc,
)
from tests.linux_kernel import LinuxKernelTestCase, parse_range_list, possible_cpus


def proc_irq_smp_affinity_list(path):
    try:
        return (path / "smp_affinity_list").read_text()
    except FileNotFoundError:
        # The smp_affinity_list file doesn't exist on !SMP or for IRQs without
        # a handler.
        return Path("/sys/devices/system/cpu/online").read_text()


class TestIrq(LinuxKernelTestCase):
    def test_for_each_irq_desc(self):
        self.assertCountEqual(
            [irq for irq, _ in for_each_irq_desc(self.prog)],
            [int(path.name) for path in Path("/sys/kernel/irq").iterdir()],
        )

    def test_irq_desc_affinity(self):
        for path in Path("/proc/irq").iterdir():
            if not path.is_dir():
                continue
            irq = int(path.name)
            self.assertCountEqual(
                list(for_each_cpu(irq_desc_affinity_mask(irq_to_desc(self.prog, irq)))),
                parse_range_list(proc_irq_smp_affinity_list(path)),
            )
            return
        self.skipTest("IRQ not found")

    def test_irq_desc_action_names(self):
        # Try to test an interrupt with one action and one with multiple actions.
        single = None
        multiple = None
        for path in Path("/sys/kernel/irq").iterdir():
            expected = (path / "actions").read_bytes().rstrip(b"\n")
            if not expected:
                continue
            irq = int(path.name)

            if b"," in expected:
                multiple = irq, expected
            else:
                single = irq, expected
            if single is not None and multiple is not None:
                break

        with self.subTest("single"):
            if single is None:
                self.skipTest("IRQ with one action not found")
            irq, expected = single
            self.assertEqual(
                b",".join(irq_desc_action_names(irq_to_desc(self.prog, irq))), expected
            )

        with self.subTest("multiple"):
            if multiple is None:
                self.skipTest("IRQ with multiple actions not found")
            irq, expected = multiple
            self.assertEqual(
                b",".join(irq_desc_action_names(irq_to_desc(self.prog, irq))), expected
            )

    def test_irq_desc_chip_name(self):
        for path in Path("/sys/kernel/irq").iterdir():
            expected = (path / "chip_name").read_bytes().rstrip(b"\n")
            if not expected:
                continue
            irq = int(path.name)
            self.assertEqual(irq_desc_chip_name(irq_to_desc(self.prog, irq)), expected)
            return
        self.skipTest("IRQ with chip name not found")

    def test_irq_desc_kstat_cpu(self):
        cpu = min(possible_cpus())
        for path in Path("/sys/kernel/irq").iterdir():
            irq = int(path.name)
            expected = int((path / "per_cpu_count").read_text().partition(",")[0])
            # Prefer a non-zero count if we can find one.
            if expected:
                break
        self.assertAlmostEqual(
            irq_desc_kstat_cpu(irq_to_desc(self.prog, irq), cpu), expected, delta=1000
        )

    @unittest.skipUnless(NORMALIZED_MACHINE_NAME == "x86_64", "machine is not x86_64")
    def test_gate_desc_func(self):
        self.assertIn(
            "int3", self.prog.symbol(gate_desc_func(self.prog["idt_table"][3])).name
        )



# Copyright (c) 2025 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn.commands import CommandArgumentError
from tests.linux_kernel.crash_commands import CrashCommandTestCase


class TestAscii(CrashCommandTestCase):
    def test_ascii(self):
        cmd = self.check_crash_command("ascii 62696c2f7273752f")
        self.assertRegex(cmd.stdout, r"(?m)^62696c2f7273752f: /usr/lib$")
        self.assertEqual(cmd.drgn_option.globals["bytestring"], b"/usr/lib")

    def test_too_large(self):
        command = "ascii fedcba9876543210f"
        self.assertRaisesRegex(
            CommandArgumentError, "too large", self.run_crash_command, command
        )
        self.assertRaisesRegex(
            ValueError, "too large", self.run_crash_command_drgn_option, command
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import re

from drgn import Symbol
from tests import slow_test
from tests.linux_kernel import skip_unless_have_test_kmod
from tests.linux_kernel.crash_commands import CrashCommandTestCase


@skip_unless_have_test_kmod
class TestSym(CrashCommandTestCase):
    @slow_test
    def test_all(self):
        cmd = self.check_crash_command("sym -l")
        self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+ init_task$")
        self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+ schedule$")
        self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+ drgn_test_function$")
        self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+ drgn_test_data$")
        self.assertNotRegex(cmd.stdout, r"(?m) \.L.*$")

        self.assertIn("prog.symbols()", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["sym"], Symbol)

    @slow_test
    def test_substring(self):
        cmd = self.check_crash_command("sym -q drgn_test")
        self.assertNotRegex(cmd.stdout, r"(?m)^[0-9a-f]+ init_task$")
        self.assertNotRegex(cmd.stdout, r"(?m)^[0-9a-f]+ schedule$")
        self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+ drgn_test_function$")
        self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+ drgn_test_data$")

        self.assertIn('"drgn_test" in sym.name', cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["sym"], Symbol)

    @slow_test
    def test_substring_multiple(self):
        cmd = self.check_crash_command("sym -q drgn_test -q init")
        self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+ init_task$")
        self.assertNotRegex(cmd.stdout, r"(?m)^[0-9a-f]+ schedule$")
        self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+ drgn_test_function$")
        self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+ drgn_test_data$")

        self.assertIn("any(substring in sym.name", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["sym"], Symbol)

    def test_name(self):
        cmd = self.check_crash_command("sym drgn_test_function")
        self.assertRegex(
            cmd.stdout, r"^[0-9a-f]+ drgn_test_function .*drgn_test\.c: [0-9]+$"
        )

        self.assertIn('prog.symbol("drgn_test_function")', cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["sym"], Symbol)
        self.assertEqual(cmd.drgn_option.globals["sym"].name, "drgn_test_function")

    def test_address(self):
        address = self.prog.symbol("drgn_test_function").address

        cmd = self.check_crash_command(f"sym {hex(address)}")
        self.assertRegex(
            cmd.stdout, r"^[0-9a-f]+ drgn_test_function .*drgn_test\.c: [0-9]+$"
        )

        self.assertIn("prog.symbol(0x", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["sym"], Symbol)
        self.assertEqual(cmd.drgn_option.globals["sym"].name, "drgn_test_function")

    def test_multiple_names(self):
        cmd = self.check_crash_command("sym init_task drgn_test_function")
        self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+ init_task$")
        self.assertRegex(
            cmd.stdout, r"(?m)^[0-9a-f]+ drgn_test_function .*drgn_test\.c: [0-9]+$"
        )

        self.assertIn('prog.symbol("drgn_test_function")', cmd.drgn_option.stdout)
        self.assertIn('prog.symbol("init_task")', cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["sym"], Symbol)
        self.assertEqual(cmd.drgn_option.globals["sym"].name, "drgn_test_function")

    @slow_test
    def test_fallback_query(self):
        # The generated --drgn output doesn't do the fallback, so it would fail
        # with a LookupError. Just compile it.
        cmd = self.check_crash_command("sym drgn_test_", mode="compile")
        self.assertNotRegex(cmd.stdout, r"(?m)^\s*[0-9a-f]+ init_task$")
        self.assertNotRegex(cmd.stdout, r"(?m)^\s*[0-9a-f]+ schedule$")
        self.assertRegex(cmd.stdout, r"(?m)^\s*[0-9a-f]+ drgn_test_function$")
        self.assertRegex(cmd.stdout, r"(?m)^\s*[0-9a-f]+ drgn_test_data$")

        self.assertIn('prog.symbol("drgn_test_")', cmd.drgn_option.stdout)

    @slow_test
    def test_next_prev(self):
        cmd = self.check_crash_command("sym -np drgn_test_function")

        lines = cmd.stdout.splitlines()
        self.assertRegex(lines[1], r"^[0-9a-f]+ drgn_test_function")

        addresses = [int(re.match(r"[0-9a-f]+", line).group(), 16) for line in lines]
        self.assertEqual(addresses, sorted(addresses))

        self.assertIsInstance(cmd.drgn_option.globals["sym"], Symbol)
        self.assertIsInstance(cmd.drgn_option.globals["prev_sym"], Symbol)
        self.assertIsInstance(cmd.drgn_option.globals["next_sym"], Symbol)

        self.assertEqual(cmd.drgn_option.globals["sym"].name, "drgn_test_function")
        self.assertLess(
            cmd.drgn_option.globals["prev_sym"].address,
            cmd.drgn_option.globals["sym"].address,
        )
        self.assertGreater(
            cmd.drgn_option.globals["next_sym"].address,
            cmd.drgn_option.globals["sym"].address,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest

from drgn import NULL, Object
from drgn.helpers.linux.xarray import (
    xa_for_each,
    xa_is_value,
    xa_is_zero,
    xa_load,
    xa_to_value,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestXArray(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if not cls.prog["drgn_test_have_xarray"]:
            raise unittest.SkipTest("kernel does not have XArray")

    def test_xa_is_zero(self):
        self.assertTrue(xa_is_zero(self.prog["drgn_test_xa_zero_entry"]))
        self.assertFalse(xa_is_zero(NULL(self.prog, "void *")))

    def test_xa_load_empty(self):
        xa = self.prog["drgn_test_xarray_empty"].address_of_()
        self.assertIdentical(xa_load(xa, 0), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 100000), NULL(self.prog, "void *"))

    def test_xa_for_each_empty(self):
        xa = self.prog["drgn_test_xarray_empty"].address_of_()
        self.assertIdentical(list(xa_for_each(xa)), [])

    def test_xa_load_one(self):
        xa = self.prog["drgn_test_xarray_one"].address_of_()
        self.assertIdentical(xa_load(xa, 0), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 665), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 666), Object(self.prog, "void *", 0xDEADB00))
        self.assertIdentical(xa_load(xa, 667), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 100000), NULL(self.prog, "void *"))

    def test_xa_for_each_one(self):
        xa = self.prog["drgn_test_xarray_one"].address_of_()
        self.assertIdentical(
            list(xa_for_each(xa)), [(666, Object(self.prog, "void *", 0xDEADB00))]
        )

    def test_xa_load_one_at_zero(self):
        xa = self.prog["drgn_test_xarray_one_at_zero"].address_of_()
        self.assertIdentical(xa_load(xa, 0), Object(self.prog, "void *", 0x1234))
        self.assertIdentical(xa_load(xa, 1), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 100000), NULL(self.prog, "void *"))

    def test_xa_for_each_one_at_zero(self):
        xa = self.prog["drgn_test_xarray_one_at_zero"].address_of_()
        self.assertIdentical(
            list(xa_for_each(xa)), [(0, Object(self.prog, "void *", 0x1234))]
        )

    def test_xa_load_sparse(self):
        xa = self.prog["drgn_test_xarray_sparse"].address_of_()
        self.assertIdentical(xa_load(xa, 0), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 1), Object(self.prog, "void *", 0x1234))
        self.assertIdentical(xa_load(xa, 2), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 0x40000000), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 0x80000000), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 0x80800000), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 0x80808000), NULL(self.prog, "void *"))
        self.assertIdentical(
            xa_load(xa, 0x80808080), Object(self.prog, "void *", 0x5678)
        )
        self.assertIdentical(xa_load(xa, 0xFFFFFFFE), NULL(self.prog, "void *"))
        self.assertIdentical(
            xa_load(xa, 0xFFFFFFFF), Object(self.prog, "void *", 0x9ABC)
        )

    def test_xa_for_each_sparse(self):
        xa = self.prog["drgn_test_xarray_sparse"].address_of_()
        self.assertIdentical(
            list(xa_for_each(xa)),
            [
                (1, Object(self.prog, "void *", 0x1234)),
                (0x80808080, Object(self.prog, "void *", 0x5678)),
                (0xFFFFFFFF, Object(self.prog, "void *", 0x9ABC)),
            ],
        )

    def test_xa_load_multi_index(self):
        xa = self.prog["drgn_test_xarray_multi_index"].address_of_()
        self.assertIdentical(xa_load(xa, 0x80807FFF), NULL(self.prog, "void *"))
        for index in range(0x80808000, 0x80808200):
            with self.subTest(index=index):
                self.assertIdentical(
                    xa_load(xa, index), Object(self.prog, "void *", 0x1234)
                )
        self.assertIdentical(xa_load(xa, 0x80808200), NULL(self.prog, "void *"))

    def test_xa_for_each_multi_index(self):
        xa = self.prog["drgn_test_xarray_multi_index"].address_of_()
        self.assertIdentical(
            list(xa_for_each(xa)), [(0x80808000, Object(self.prog, "void *", 0x1234))]
        )

    def test_xa_load_zero_entry(self):
        xa = self.prog["drgn_test_xarray_zero_entry"].address_of_()
        self.assertIdentical(xa_load(xa, 0), NULL(self.prog, "void *"))
        self.assertIdentical(xa_load(xa, 666), NULL(self.prog, "void *"))
        self.assertTrue(xa_is_zero(xa_load(xa, 666, advanced=True)))
        self.assertIdentical(xa_load(xa, 2), NULL(self.prog, "void *"))

    def test_xa_for_each_zero_entry(self):
        xa = self.prog["drgn_test_xarray_zero_entry"].address_of_()
        self.assertIdentical(list(xa_for_each(xa)), [])

        entries = list(xa_for_each(xa, advanced=True))
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0][0], 666)
        self.assertTrue(xa_is_zero(entries[0][1]))

    def test_xa_load_zero_entry_at_zero(self):
        xa = self.prog["drgn_test_xarray_zero_entry_at_zero"].address_of_()
        self.assertIdentical(xa_load(xa, 0), NULL(self.prog, "void *"))
        self.assertTrue(xa_is_zero(xa_load(xa, 0, advanced=True)))
        self.assertIdentical(xa_load(xa, 1), NULL(self.prog, "void *"))

    def test_xa_for_each_zero_entry_at_zero(self):
        xa = self.prog["drgn_test_xarray_zero_entry_at_zero"].address_of_()
        self.assertIdentical(list(xa_for_each(xa)), [])

        entries = list(xa_for_each(xa, advanced=True))
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0][0], 0)
        self.assertTrue(xa_is_zero(entries[0][1]))

    def test_xa_is_value(self):
        self.assertTrue(
            xa_is_value(xa_load(self.prog["drgn_test_xarray_value"].address_of_(), 0))
        )
        self.assertFalse(
            xa_is_value(
                xa_load(self.prog["drgn_test_xarray_one_at_zero"].address_of_(), 0)
            )
        )

    def test_xa_to_value(self):
        self.assertIdentical(
            xa_to_value(xa_load(self.prog["drgn_test_xarray_value"].address_of_(), 0)),
            Object(self.prog, "unsigned long", 1337),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
# Generated by scripts/gen_dwarf_constants.py.

import enum


class DW_ACCESS(enum.IntEnum):
    public = 0x1
    protected = 0x2
    private = 0x3


class DW_ADDR(enum.IntEnum):
    none = 0x0
    TI_PTR8 = 0x8
    TI_PTR16 = 0x10
    TI_PTR22 = 0x16
    TI_PTR23 = 0x17
    TI_PTR24 = 0x18
    TI_PTR32 = 0x20


class DW_AT(enum.IntEnum):
    sibling = 0x1
    location = 0x2
    name = 0x3  # type: ignore
    ordering = 0x9
    subscr_data = 0xA
    byte_size = 0xB
    bit_offset = 0xC
    bit_size = 0xD
    element_list = 0xF
    stmt_list = 0x10
    low_pc = 0x11
    high_pc = 0x12
    language = 0x13
    member = 0x14
    discr = 0x15
    discr_value = 0x16
    visibility = 0x17
    import_ = 0x18
    string_length = 0x19
    common_reference = 0x1A
    comp_dir = 0x1B
    const_value = 0x1C
    containing_type = 0x1D
    default_value = 0x1E
    inline = 0x20
    is_optional = 0x21
    lower_bound = 0x22
    producer = 0x25
    prototyped = 0x27
    return_addr = 0x2A
    start_scope = 0x2C
    bit_stride = 0x2E
    stride_size = 0x2E
    upper_bound = 0x2F
    abstract_origin = 0x31
    accessibility = 0x32
    address_class = 0x33
    artificial = 0x34
    base_types = 0x35
    calling_convention = 0x36
    count = 0x37
    data_member_location = 0x38
    decl_column = 0x39
    decl_file = 0x3A
    decl_line = 0x3B
    declaration = 0x3C
    discr_list = 0x3D
    encoding = 0x3E
    external = 0x3F
    frame_base = 0x40
    friend = 0x41
    identifier_case = 0x42
    macro_info = 0x43
    namelist_item = 0x44
    priority = 0x45
    segment = 0x46
    specification = 0x47
    static_link = 0x48
    type = 0x49
    use_location = 0x4A
    variable_parameter = 0x4B
    virtuality = 0x4C
    vtable_elem_location = 0x4D
    allocated = 0x4E
    associated = 0x4F
    data_location = 0x50
    byte_stride = 0x51
    entry_pc = 0x52
    use_UTF8 = 0x53
    extension = 0x54
    ranges = 0x55
    trampoline = 0x56
    call_column = 0x57
    call_file = 0x58
    call_line = 0x59
    description = 0x5A
    binary_scale = 0x5B
    decimal_scale = 0x5C
    small = 0x5D
    decimal_sign = 0x5E
    digit_count = 0x5F
    picture_string = 0x60
    mutable = 0x61
    threads_scaled = 0x62
    explicit = 0x63
    object_pointer = 0x64
    endianity = 0x65
    elemental = 0x66
    pure = 0x67
    recursive = 0x68
    signature = 0x69
    main_subprogram = 0x6A
    data_bit_offset = 0x6B
    const_expr = 0x6C
    enum_class = 0x6D
    linkage_name = 0x6E
    string_length_bit_size = 0x6F
    string_length_byte_size = 0x70
    rank = 0x71
    str_offsets_base = 0x72
    addr_base = 0x73
    rnglists_base = 0x74
    dwo_id = 0x75
    dwo_name = 0x76
    reference = 0x77
    rvalue_reference = 0x78
    macros = 0x79
    call_all_calls = 0x7A
    call_all_source_calls = 0x7B
    call_all_tail_calls = 0x7C
    call_return_pc = 0x7D
    call_value = 0x7E
    call_origin = 0x7F
    call_parameter = 0x80
    call_pc = 0x81
    call_tail_call = 0x82
    call_target = 0x83
    call_target_clobbered = 0x84
    call_data_location = 0x85
    call_data_value = 0x86
    noreturn = 0x87
    alignment = 0x88
    export_symbols = 0x89
    deleted = 0x8A
    defaulted = 0x8B
    loclists_base = 0x8C
    ghs_namespace_alias = 0x806
    ghs_using_namespace = 0x807
    ghs_using_declaration = 0x808
    HP_block_index = 0x2000
    lo_user = 0x2000
    TI_veneer = 0x2000
    MIPS_fde = 0x2001
    TI_symbol_name = 0x2001
    MIPS_loop_begin = 0x2002
    MIPS_tail_loop_begin = 0x2003
    MIPS_epilog_begin = 0x2004
    MIPS_loop_unroll_factor = 0x2005
    MIPS_software_pipeline_depth = 0x2006
    MIPS_linkage_name = 0x2007
    MIPS_stride = 0x2008
    MIPS_abstract_name = 0x2009
    MIPS_clone_origin = 0x200A
    MIPS_has_inlines = 0x200B
    TI_version = 0x200B
    MIPS_stride_byte = 0x200C
    TI_asm = 0x200C
    MIPS_stride_elem = 0x200D
    MIPS_ptr_dopetype = 0x200E
    TI_skeletal = 0x200E
    MIPS_allocatable_dopetype = 0x200F
    MIPS_assumed_shape_dopetype = 0x2010
    MIPS_assumed_size = 0x2011
    TI_interrupt = 0x2011
    HP_unmodifiable = 0x2001
    HP_prologue = 0x2005
    HP_epilogue = 0x2008
    HP_actuals_stmt_list = 0x2010
    HP_proc_per_section = 0x2011
    HP_raw_data_ptr = 0x2012
    HP_pass_by_reference = 0x2013
    HP_opt_level = 0x2014
    HP_prof_version_id = 0x2015
    HP_opt_flags = 0x2016
    HP_cold_region_low_pc = 0x2017
    HP_cold_region_high_pc = 0x2018
    HP_all_variables_modifiable = 0x2019
    HP_linkage_name = 0x201A
    HP_prof_flags = 0x201B
    HP_unit_name = 0x201F
    HP_unit_size = 0x2020
    HP_widened_byte_size = 0x2021
    HP_definition_points = 0x2022
    HP_default_location = 0x2023
    HP_is_result_param = 0x2029
    CPQ_discontig_ranges = 0x2001
    CPQ_semantic_events = 0x2002
    CPQ_split_lifetimes_var = 0x2003
    CPQ_split_lifetimes_rtn = 0x2004
    CPQ_prologue_length = 0x2005
    ghs_mangled = 0x2007
    ghs_rsm = 0x2083
    ghs_frsm = 0x2085
    ghs_frames = 0x2086
    ghs_rso = 0x2087
    ghs_subcpu = 0x2092
    ghs_lbrace_line = 0x2093
    INTEL_other_endian = 0x2026
    sf_names = 0x2101
    src_info = 0x2102
    mac_info = 0x2103
    src_coords = 0x2104
    body_begin = 0x2105
    body_end = 0x2106
    GNU_vector = 0x2107
    GNU_guarded_by = 0x2108
    GNU_pt_guarded_by = 0x2109
    GNU_guarded = 0x210A
    GNU_pt_guarded = 0x210B
    GNU_locks_excluded = 0x210C
    GNU_exclusive_locks_required = 0x210D
    GNU_shared_locks_required = 0x210E
    GNU_odr_signature = 0x210F
    GNU_template_name = 0x2110
    GNU_call_site_value = 0x2111
    GNU_call_site_data_value = 0x2112
    GNU_call_site_target = 0x2113
    GNU_call_site_target_clobbered = 0x2114
    GNU_tail_call = 0x2115
    GNU_all_tail_call_sites = 0x2116
    GNU_all_call_sites = 0x2117
    GNU_all_source_call_sites = 0x2118
    GNU_macros = 0x2119
    GNU_deleted = 0x211A
    GNU_dwo_name = 0x2130
    GNU_dwo_id = 0x2131
    GNU_ranges_base = 0x2132
    GNU_addr_base = 0x2133
    GNU_pubnames = 0x2134
    GNU_pubtypes = 0x2135
    GNU_discriminator = 0x2136
    GNU_locviews = 0x2137
    GNU_entry_view = 0x2138
    SUN_template = 0x2201
    VMS_rtnbeg_pd_address = 0x2201
    SUN_alignment = 0x2202
    SUN_vtable = 0x2203
    SUN_count_guarantee = 0x2204
    SUN_command_line = 0x2205
    SUN_vbase = 0x2206
    SUN_compile_options = 0x2207
    SUN_language = 0x2208
    SUN_browser_file = 0x2209
    SUN_vtable_abi = 0x2210
    SUN_func_offsets = 0x2211
    SUN_cf_kind = 0x2212
    SUN_vtable_index = 0x2213
    SUN_omp_tpriv_addr = 0x2214
    SUN_omp_child_func = 0x2215
    SUN_func_offset = 0x2216
    SUN_memop_type_ref = 0x2217
    SUN_profile_id = 0x2218
    SUN_memop_signature = 0x2219
    SUN_obj_dir = 0x2220
    SUN_obj_file = 0x2221
    SUN_original_name = 0x2222
    SUN_hwcprof_signature = 0x2223
    SUN_amd64_parmdump = 0x2224
    SUN_part_link_name = 0x2225
    SUN_link_name = 0x2226
    SUN_pass_with_const = 0x2227
    SUN_return_with_const = 0x2228
    SUN_import_by_name = 0x2229
    SUN_f90_pointer = 0x222A
    SUN_pass_by_ref = 0x222B
    SUN_f90_allocatable = 0x222C
    SUN_f90_assumed_shape_array = 0x222D
    SUN_c_vla = 0x222E
    SUN_return_value_ptr = 0x2230
    SUN_dtor_start = 0x2231
    SUN_dtor_length = 0x2232
    SUN_dtor_state_initial = 0x2233
    SUN_dtor_state_final = 0x2234
    SUN_dtor_state_deltas = 0x2235
    SUN_import_by_lname = 0x2236
    SUN_f90_use_only = 0x2237
    SUN_namelist_spec = 0x2238
    SUN_is_omp_child_func = 0x2239
    SUN_fortran_main_alias = 0x223A
    SUN_fortran_based = 0x223B
    ALTIUM_loclist = 0x2300
    use_GNAT_descriptive_type = 0x2301
    GNAT_descriptive_type = 0x2302
    GNU_numerator = 0x2303
    GNU_denominator = 0x2304
    GNU_bias = 0x2305
    go_kind = 0x2900
    go_key = 0x2901
    go_elem = 0x2902
    go_embedded_field = 0x2903
    go_runtime_type = 0x2904
    upc_threads_scaled = 0x3210
    IBM_wsa_addr = 0x393E
    IBM_home_location = 0x393F
    IBM_alt_srcview = 0x3940
    PGI_lbase = 0x3A00
    PGI_soffset = 0x3A01
    PGI_lstride = 0x3A02
    BORLAND_property_read = 0x3B11
    BORLAND_property_write = 0x3B12
    BORLAND_property_implements = 0x3B13
    BORLAND_property_index = 0x3B14
    BORLAND_property_default = 0x3B15
    BORLAND_Delphi_unit = 0x3B20
    BORLAND_Delphi_class = 0x3B21
    BORLAND_Delphi_record = 0x3B22
    BORLAND_Delphi_metaclass = 0x3B23
    BORLAND_Delphi_constructor = 0x3B24
    BORLAND_Delphi_destructor = 0x3B25
    BORLAND_Delphi_anonymous_method = 0x3B26
    BORLAND_Delphi_interface = 0x3B27
    BORLAND_Delphi_ABI = 0x3B28
    BORLAND_Delphi_frameptr = 0x3B30
    BORLAND_closure = 0x3B31
    LLVM_include_path = 0x3E00
    LLVM_config_macros = 0x3E01
    LLVM_sysroot = 0x3E02
    LLVM_tag_offset = 0x3E03
    LLVM_apinotes = 0x3E07
    LLVM_active_lane = 0x3E08
    LLVM_augmentation = 0x3E09
    LLVM_lanes = 0x3E0A
    LLVM_lane_pc = 0x3E0B
    LLVM_vector_size = 0x3E0C
    APPLE_optimized = 0x3FE1
    APPLE_flags = 0x3FE2
    APPLE_isa = 0x3FE3
    APPLE_block = 0x3FE4
    APPLE_major_runtime_vers = 0x3FE5
    APPLE_runtime_class = 0x3FE6
    APPLE_omit_frame_ptr = 0x3FE7
    APPLE_property_name = 0x3FE8
    APPLE_property_getter = 0x3FE9
    APPLE_property_setter = 0x3FEA
    APPLE_property_attribute = 0x3FEB
    APPLE_objc_complete_type = 0x3FEC
    APPLE_property = 0x3FED
    APPLE_objc_direct = 0x3FEE
    APPLE_sdk = 0x3FEF
    APPLE_origin = 0x3FF0
    hi_user = 0x3FFF


class DW_ATE(enum.IntEnum):
    address = 0x1
    boolean = 0x2
    complex_float = 0x3
    float = 0x4
    signed = 0x5
    signed_char = 0x6
    unsigned = 0x7
    unsigned_char = 0x8
    imaginary_float = 0x9
    packed_decimal = 0xA
    numeric_string = 0xB
    edited = 0xC
    signed_fixed = 0xD
    unsigned_fixed = 0xE
    decimal_float = 0xF
    UTF = 0x10
    UCS = 0x11
    ASCII = 0x12
    ALTIUM_fract = 0x80
    lo_user = 0x80
    ALTIUM_accum = 0x81
    HP_float80 = 0x80
    HP_complex_float80 = 0x81
    HP_float128 = 0x82
    HP_complex_float128 = 0x83
    HP_floathpintel = 0x84
    HP_imaginary_float80 = 0x85
    HP_imaginary_float128 = 0x86
    HP_VAX_float = 0x88
    HP_VAX_float_d = 0x89
    HP_packed_decimal = 0x8A
    HP_zoned_decimal = 0x8B
    HP_edited = 0x8C
    HP_signed_fixed = 0x8D
    HP_unsigned_fixed = 0x8E
    HP_VAX_complex_float = 0x8F
    HP_VAX_complex_float_d = 0x90
    SUN_interval_float = 0x91
    SUN_imaginary_float = 0x92
    hi_user = 0xFF


class DW_CC(enum.IntEnum):
    normal = 0x1
    program = 0x2
    nocall = 0x3
    pass_by_reference = 0x4
    pass_by_value = 0x5
    GNU_renesas_sh = 0x40
    lo_user = 0x40
    GNU_borland_fastcall_i386 = 0x41
    ALTIUM_interrupt = 0x65
    ALTIUM_near_system_stack = 0x66
    ALTIUM_near_user_stack = 0x67
    ALTIUM_huge_user_stack = 0x68
    GNU_BORLAND_safecall = 0xB0
    GNU_BORLAND_stdcall = 0xB1
    GNU_BORLAND_pascal = 0xB2
    GNU_BORLAND_msfastcall = 0xB3
    GNU_BORLAND_msreturn = 0xB4
    GNU_BORLAND_thiscall = 0xB5
    GNU_BORLAND_fastcall = 0xB6
    LLVM_vectorcall = 0xC0
    LLVM_Win64 = 0xC1
    LLVM_X86_64SysV = 0xC2
    LLVM_AAPCS = 0xC3
    LLVM_AAPCS_VFP = 0xC4
    LLVM_IntelOclBicc = 0xC5
    LLVM_SpirFunction = 0xC6
    LLVM_OpenCLKernel = 0xC7
    LLVM_Swift = 0xC8
    LLVM_PreserveMost = 0xC9
    LLVM_PreserveAll = 0xCA
    LLVM_X86RegCall = 0xCB
    GDB_IBM_OpenCL = 0xFF
    hi_user = 0xFF


class DW_CFA(enum.IntEnum):
    advance_loc = 0x40
    offset = 0x80
    restore = 0xC0
    nop = 0x0
    set_loc = 0x1
    advance_loc1 = 0x2
    advance_loc2 = 0x3
    advance_loc4 = 0x4
    offset_extended = 0x5
    restore_extended = 0x6
    undefined = 0x7
    same_value = 0x8
    register = 0x9
    remember_state = 0xA
    restore_state = 0xB
    def_cfa = 0xC
    def_cfa_register = 0xD
    def_cfa_offset = 0xE
    def_cfa_expression = 0xF
    expression = 0x10
    offset_extended_sf = 0x11
    def_cfa_sf = 0x12
    def_cfa_offset_sf = 0x13
    val_offset = 0x14
    val_offset_sf = 0x15
    val_expression = 0x16
    TI_soffset_extended = 0x1C
    lo_user = 0x1C
    MIPS_advance_loc8 = 0x1D
    TI_def_cfa_soffset = 0x1D
    GNU_window_save = 0x2D
    AARCH64_negate_ra_state = 0x2D
    GNU_args_size = 0x2E
    GNU_negative_offset_extended = 0x2F
    LLVM_def_aspace_cfa = 0x30
    LLVM_def_aspace_cfa_sf = 0x31
    METAWARE_info = 0x34
    hi_user = 0x3F


class DW_CHILDREN(enum.IntEnum):
    no = 0x0
    yes = 0x1


class DW_DEFAULTED(enum.IntEnum):
    no = 0x0
    in_class = 0x1
    out_of_class = 0x2


class DW_DS(enum.IntEnum):
    unsigned = 0x1
    leading_overpunch = 0x2
    trailing_overpunch = 0x3
    leading_separate = 0x4
    trailing_separate = 0x5


class DW_DSC(enum.IntEnum):
    label = 0x0
    range = 0x1


class DW_EH_PE(enum.IntEnum):
    absptr = 0x0
    uleb128 = 0x1
    udata2 = 0x2
    udata4 = 0x3
    udata8 = 0x4
    sleb128 = 0x9
    sdata2 = 0xA
    sdata4 = 0xB
    sdata8 = 0xC
    signed = 0x8
    pcrel = 0x10
    textrel = 0x20
    datarel = 0x30
    funcrel = 0x40
    aligned = 0x50
    indirect = 0x80
    omit = 0xFF


class DW_END(enum.IntEnum):
    default = 0x0
    big = 0x1
    little = 0x2
    lo_user = 0x40
    hi_user = 0xFF


class DW_FORM(enum.IntEnum):
    addr = 0x1
    block2 = 0x3
    block4 = 0x4
    data2 = 0x5
    data4 = 0x6
    data8 = 0x7
    string = 0x8
    block = 0x9
    block1 = 0xA
    data1 = 0xB
    flag = 0xC
    sdata = 0xD
    strp = 0xE
    udata = 0xF
    ref_addr = 0x10
    ref1 = 0x11
    ref2 = 0x12
    ref4 = 0x13
    ref8 = 0x14
    ref_udata = 0x15
    indirect = 0x16
    sec_offset = 0x17
    exprloc = 0x18
    flag_present = 0x19
    strx = 0x1A
    addrx = 0x1B
    ref_sup4 = 0x1C
    strp_sup = 0x1D
    data16 = 0x1E
    line_strp = 0x1F
    ref_sig8 = 0x20
    implicit_const = 0x21
    loclistx = 0x22
    rnglistx = 0x23
    ref_sup8 = 0x24
    strx1 = 0x25
    strx2 = 0x26
    strx3 = 0x27
    strx4 = 0x28
    addrx1 = 0x29
    addrx2 = 0x2A
    addrx3 = 0x2B
    addrx4 = 0x2C
    GNU_addr_index = 0x1F01
    GNU_str_index = 0x1F02
    GNU_ref_alt = 0x1F20
    GNU_strp_alt = 0x1F21
    LLVM_addrx_offset = 0x2001


class DW_ID(enum.IntEnum):
    case_sensitive = 0x0
    up_case = 0x1
    down_case = 0x2
    case_insensitive = 0x3


class DW_IDX(enum.IntEnum):
    compile_unit = 0x1
    type_unit = 0x2
    die_offset = 0x3
    parent = 0x4
    type_hash = 0x5
    GNU_internal = 0x2000
    lo_user = 0x2000
    GNU_external = 0x2001
    GNU_main = 0x2002
    GNU_language = 0x2003
    GNU_linkage_name = 0x2004
    hi_user = 0x3FFF


class DW_INL(enum.IntEnum):
    not_inlined = 0x0
    inlined = 0x1
    declared_not_inlined = 0x2
    declared_inlined = 0x3


class DW_LANG(enum.IntEnum):
    C89 = 0x1
    C = 0x2
    Ada83 = 0x3
    C_plus_plus = 0x4
    Cobol74 = 0x5
    Cobol85 = 0x6
    Fortran77 = 0x7
    Fortran90 = 0x8
    Pascal83 = 0x9
    Modula2 = 0xA
    Java = 0xB
    C99 = 0xC
    Ada95 = 0xD
    Fortran95 = 0xE
    PLI = 0xF
    ObjC = 0x10
    ObjC_plus_plus = 0x11
    UPC = 0x12
    D = 0x13
    Python = 0x14
    OpenCL = 0x15
    Go = 0x16
    Modula3 = 0x17
    Haskell = 0x18
    C_plus_plus_03 = 0x19
    C_plus_plus_11 = 0x1A
    OCaml = 0x1B
    Rust = 0x1C
    C11 = 0x1D
    Swift = 0x1E
    Julia = 0x1F
    Dylan = 0x20
    C_plus_plus_14 = 0x21
    Fortran03 = 0x22
    Fortran08 = 0x23
    RenderScript = 0x24
    BLISS = 0x25
    Kotlin = 0x26
    Zig = 0x27
    Crystal = 0x28
    C_plus_plus_17 = 0x2A
    C_plus_plus_20 = 0x2B
    C17 = 0x2C
    Fortran18 = 0x2D
    Ada2005 = 0x2E
    Ada2012 = 0x2F
    HIP = 0x30
    Assembly = 0x31
    C_sharp = 0x32
    Mojo = 0x33
    GLSL = 0x34
    GLSL_ES = 0x35
    HLSL = 0x36
    OpenCL_CPP = 0x37
    CPP_for_OpenCL = 0x38
    SYCL = 0x39
    Ruby = 0x40
    Move = 0x41
    Hylo = 0x42
    lo_user = 0x8000
    Mips_Assembler = 0x8001
    Upc = 0x8765
    GOOGLE_RenderScript = 0x8E57
    ALTIUM_Assembler = 0x9101
    BORLAND_Delphi = 0xB000
    SUN_Assembler = 0x9001
    hi_user = 0xFFFF


class DW_LLE(enum.IntEnum):
    end_of_list = 0x0
    base_addressx = 0x1
    startx_endx = 0x2
    startx_length = 0x3
    offset_pair = 0x4
    default_location = 0x5
    base_address = 0x6
    start_end = 0x7
    start_length = 0x8


class DW_LNCT(enum.IntEnum):
    path = 0x1
    directory_index = 0x2
    timestamp = 0x3
    size = 0x4
    MD5 = 0x5
    GNU_subprogram_name = 0x6
    GNU_decl_file = 0x7
    GNU_decl_line = 0x8
    lo_user = 0x2000
    LLVM_source = 0x2001
    LLVM_is_MD5 = 0x2002
    hi_user = 0x3FFF


class DW_LNE(enum.IntEnum):
    end_sequence = 0x1
    set_address = 0x2
    define_file = 0x3
    set_discriminator = 0x4
    HP_negate_is_UV_update = 0x11
    HP_push_context = 0x12
    HP_pop_context = 0x13
    HP_set_file_line_column = 0x14
    HP_set_routine_name = 0x15
    HP_set_sequence = 0x16
    HP_negate_post_semantics = 0x17
    HP_negate_function_exit = 0x18
    HP_negate_front_end_logical = 0x19
    HP_define_proc = 0x20
    HP_source_file_correlation = 0x80
    lo_user = 0x80
    hi_user = 0xFF


class DW_LNS(enum.IntEnum):
    copy = 0x1
    advance_pc = 0x2
    advance_line = 0x3
    set_file = 0x4
    set_column = 0x5
    negate_stmt = 0x6
    set_basic_block = 0x7
    const_add_pc = 0x8
    fixed_advance_pc = 0x9
    set_prologue_end = 0xA
    set_epilogue_begin = 0xB
    set_isa = 0xC
    set_address_from_logical = 0xD
    set_subprogram = 0xD
    inlined_call = 0xE
    pop_context = 0xF


class DW_MACINFO(enum.IntEnum):
    define = 0x1
    undef = 0x2
    start_file = 0x3
    end_file = 0x4
    vendor_ext = 0xFF


class DW_MACRO(enum.IntEnum):
    define = 0x1
    undef = 0x2
    start_file = 0x3
    end_file = 0x4
    define_strp = 0x5
    undef_strp = 0x6
    import_ = 0x7
    define_sup = 0x8
    undef_sup = 0x9
    import_sup = 0xA
    define_strx = 0xB
    undef_strx = 0xC
    lo_user = 0xE0
    hi_user = 0xFF


class DW_OP(enum.IntEnum):
    addr = 0x3
    deref = 0x6
    const1u = 0x8
    const1s = 0x9
    const2u = 0xA
    const2s = 0xB
    const4u = 0xC
    const4s = 0xD
    const8u = 0xE
    const8s = 0xF
    constu = 0x10
    consts = 0x11
    dup = 0x12
    drop = 0x13
    over = 0x14
    pick = 0x15
    swap = 0x16
    rot = 0x17
    xderef = 0x18
    abs = 0x19
    and_ = 0x1A
    div = 0x1B
    minus = 0x1C
    mod = 0x1D
    mul = 0x1E
    neg = 0x1F
    not_ = 0x20
    or_ = 0x21
    plus = 0x22
    plus_uconst = 0x23
    shl = 0x24
    shr = 0x25
    shra = 0x26
    xor = 0x27
    bra = 0x28
    eq = 0x29
    ge = 0x2A
    gt = 0x2B
    le = 0x2C
    lt = 0x2D
    ne = 0x2E
    skip = 0x2F
    lit0 = 0x30
    lit1 = 0x31
    lit2 = 0x32
    lit3 = 0x33
    lit4 = 0x34
    lit5 = 0x35
    lit6 = 0x36
    lit7 = 0x37
    lit8 = 0x38
    lit9 = 0x39
    lit10 = 0x3A
    lit11 = 0x3B
    lit12 = 0x3C
    lit13 = 0x3D
    lit14 = 0x3E
    lit15 = 0x3F
    lit16 = 0x40
    lit17 = 0x41
    lit18 = 0x42
    lit19 = 0x43
    lit20 = 0x44
    lit21 = 0x45
    lit22 = 0x46
    lit23 = 0x47
    lit24 = 0x48
    lit25 = 0x49
    lit26 = 0x4A
    lit27 = 0x4B
    lit28 = 0x4C
    lit29 = 0x4D
    lit30 = 0x4E
    lit31 = 0x4F
    reg0 = 0x50
    reg1 = 0x51
    reg2 = 0x52
    reg3 = 0x53
    reg4 = 0x54
    reg5 = 0x55
    reg6 = 0x56
    reg7 = 0x57
    reg8 = 0x58
    reg9 = 0x59
    reg10 = 0x5A
    reg11 = 0x5B
    reg12 = 0x5C
    reg13 = 0x5D
    reg14 = 0x5E
    reg15 = 0x5F
    reg16 = 0x60
    reg17 = 0x61
    reg18 = 0x62
    reg19 = 0x63
    reg20 = 0x64
    reg21 = 0x65
    reg22 = 0x66
    reg23 = 0x67
    reg24 = 0x68
    reg25 = 0x69
    reg26 = 0x6A
    reg27 = 0x6B
    reg28 = 0x6C
    reg29 = 0x6D
    reg30 = 0x6E
    reg31 = 0x6F
    breg0 = 0x70
    breg1 = 0x71
    breg2 = 0x72
    breg3 = 0x73
    breg4 = 0x74
    breg5 = 0x75
    breg6 = 0x76
    breg7 = 0x77
    breg8 = 0x78
    breg9 = 0x79
    breg10 = 0x7A
    breg11 = 0x7B
    breg12 = 0x7C
    breg13 = 0x7D
    breg14 = 0x7E
    breg15 = 0x7F
    breg16 = 0x80
    breg17 = 0x81
    breg18 = 0x82
    breg19 = 0x83
    breg20 = 0x84
    breg21 = 0x85
    breg22 = 0x86
    breg23 = 0x87
    breg24 = 0x88
    breg25 = 0x89
    breg26 = 0x8A
    breg27 = 0x8B
    breg28 = 0x8C
    breg29 = 0x8D
    breg30 = 0x8E
    breg31 = 0x8F
    regx = 0x90
    fbreg = 0x91
    bregx = 0x92
    piece = 0x93
    deref_size = 0x94
    xderef_size = 0x95
    nop = 0x96
    push_object_address = 0x97
    call2 = 0x98
    call4 = 0x99
    call_ref = 0x9A
    form_tls_address = 0x9B
    call_frame_cfa = 0x9C
    bit_piece = 0x9D
    implicit_value = 0x9E
    stack_value = 0x9F
    implicit_pointer = 0xA0
    addrx = 0xA1
    constx = 0xA2
    entry_value = 0xA3
    const_type = 0xA4
    regval_type = 0xA5
    deref_type = 0xA6
    xderef_type = 0xA7
    convert = 0xA8
    reinterpret = 0xA9
    GNU_push_tls_address = 0xE0
    WASM_location = 0xED
    WASM_location_int = 0xEE
    lo_user = 0xE0
    LLVM_form_aspace_address = 0xE1
    LLVM_push_lane = 0xE2
    LLVM_offset = 0xE3
    LLVM_offset_uconst = 0xE4
    LLVM_bit_offset = 0xE5
    LLVM_call_frame_entry_reg = 0xE6
    LLVM_undefined = 0xE7
    LLVM_aspace_bregx = 0xE8
    LLVM_aspace_implicit_pointer = 0xE9
    LLVM_piece_end = 0xEA
    LLVM_extend = 0xEB
    LLVM_select_bit_piece = 0xEC
    HP_unknown = 0xE0
    HP_is_value = 0xE1
    HP_fltconst4 = 0xE2
    HP_fltconst8 = 0xE3
    HP_mod_range = 0xE4
    HP_unmod_range = 0xE5
    HP_tls = 0xE6
    INTEL_bit_piece = 0xE8
    GNU_uninit = 0xF0
    APPLE_uninit = 0xF0
    GNU_encoded_addr = 0xF1
    GNU_implicit_pointer = 0xF2
    GNU_entry_value = 0xF3
    GNU_const_type = 0xF4
    GNU_regval_type = 0xF5
    GNU_deref_type = 0xF6
    GNU_convert = 0xF7
    GNU_reinterpret = 0xF9
    GNU_parameter_ref = 0xFA
    GNU_addr_index = 0xFB
    GNU_const_index = 0xFC
    GNU_variable_value = 0xFD
    PGI_omp_thread_num = 0xF8
    hi_user = 0xFF


class DW_ORD(enum.IntEnum):
    row_major = 0x0
    col_major = 0x1


class DW_RLE(enum.IntEnum):
    end_of_list = 0x0
    base_addressx = 0x1
    startx_endx = 0x2
    startx_length = 0x3
    offset_pair = 0x4
    base_address = 0x5
    start_end = 0x6
    start_length = 0x7


class DW_SECT(enum.IntEnum):
    INFO = 0x1
    TYPES = 0x2
    ABBREV = 0x3
    LINE = 0x4
    LOCLISTS = 0x5
    STR_OFFSETS = 0x6
    MACRO = 0x7
    RNGLISTS = 0x8


class DW_TAG(enum.IntEnum):
    array_type = 0x1
    class_type = 0x2
    entry_point = 0x3
    enumeration_type = 0x4
    formal_parameter = 0x5
    imported_declaration = 0x8
    label = 0xA
    lexical_block = 0xB
    member = 0xD
    pointer_type = 0xF
    reference_type = 0x10
    compile_unit = 0x11
    string_type = 0x12
    structure_type = 0x13
    subroutine_type = 0x15
    typedef = 0x16
    union_type = 0x17
    unspecified_parameters = 0x18
    variant = 0x19
    common_block = 0x1A
    common_inclusion = 0x1B
    inheritance = 0x1C
    inlined_subroutine = 0x1D
    module = 0x1E
    ptr_to_member_type = 0x1F
    set_type = 0x20
    subrange_type = 0x21
    with_stmt = 0x22
    access_declaration = 0x23
    base_type = 0x24
    catch_block = 0x25
    const_type = 0x26
    constant = 0x27
    enumerator = 0x28
    file_type = 0x29
    friend = 0x2A
    namelist = 0x2B
    namelist_item = 0x2C
    packed_type = 0x2D
    subprogram = 0x2E
    template_type_parameter = 0x2F
    template_value_parameter = 0x30
    thrown_type = 0x31
    try_block = 0x32
    variant_part = 0x33
    variable = 0x34
    volatile_type = 0x35
    dwarf_procedure = 0x36
    restrict_type = 0x37
    interface_type = 0x38
    namespace = 0x39
    imported_module = 0x3A
    unspecified_type = 0x3B
    partial_unit = 0x3C
    imported_unit = 0x3D
    mutable_type = 0x3E
    condition = 0x3F
    shared_type = 0x40
    type_unit = 0x41
    rvalue_reference_type = 0x42
    template_alias = 0x43
    coarray_type = 0x44
    generic_subrange = 0x45
    dynamic_type = 0x46
    atomic_type = 0x47
    call_site = 0x48
    call_site_parameter = 0x49
    skeleton_unit = 0x4A
    immutable_type = 0x4B
    TI_far_type = 0x4080
    lo_user = 0x4080
    MIPS_loop = 0x4081
    TI_near_type = 0x4081
    TI_assign_register = 0x4082
    TI_ioport_type = 0x4083
    TI_restrict_type = 0x4084
    TI_onchip_type = 0x4085
    HP_array_descriptor = 0x4090
    format_label = 0x4101
    function_template = 0x4102
    class_template = 0x4103
    GNU_BINCL = 0x4104
    GNU_EINCL = 0x4105
    GNU_template_template_param = 0x4106
    GNU_template_parameter_pack = 0x4107
    GNU_formal_parameter_pack = 0x4108
    GNU_call_site = 0x4109
    GNU_call_site_parameter = 0x410A
    SUN_function_template = 0x4201
    SUN_class_template = 0x4202
    SUN_struct_template = 0x4203
    SUN_union_template = 0x4204
    SUN_indirect_inheritance = 0x4205
    SUN_codeflags = 0x4206
    SUN_memop_info = 0x4207
    SUN_omp_child_func = 0x4208
    SUN_rtti_descriptor = 0x4209
    SUN_dtor_info = 0x420A
    SUN_dtor = 0x420B
    SUN_f90_interface = 0x420C
    SUN_fortran_vax_structure = 0x420D
    SUN_hi = 0x42FF
    ALTIUM_circ_type = 0x5101
    ALTIUM_mwa_circ_type = 0x5102
    ALTIUM_rev_carry_type = 0x5103
    ALTIUM_rom = 0x5111
    LLVM_annotation = 0x6000
    ghs_namespace = 0x8004
    ghs_using_namespace = 0x8005
    ghs_using_declaration = 0x8006
    ghs_template_templ_param = 0x8007
    upc_shared_type = 0x8765
    upc_strict_type = 0x8766
    upc_relaxed_type = 0x8767
    PGI_kanji_type = 0xA000
    PGI_interface_block = 0xA020
    BORLAND_property = 0xB000
    BORLAND_Delphi_string = 0xB001
    BORLAND_Delphi_dynamic_array = 0xB002
    BORLAND_Delphi_set = 0xB003
    BORLAND_Delphi_variant = 0xB004
    hi_user = 0xFFFF


class DW_UT(enum.IntEnum):
    compile = 0x1
    type = 0x2
    partial = 0x3
    skeleton = 0x4
    split_compile = 0x5
    split_type = 0x6
    lo_user = 0x80
    hi_user = 0xFF


class DW_VIRTUALITY(enum.IntEnum):
    none = 0x0
    virtual = 0x1
    pure_virtual = 0x2


class DW_VIS(enum.IntEnum):
    local = 0x1
    exported = 0x2
    qualified = 0x3



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
from tests import TestCase
from tests.libdrgn import deserialize_bits, serialize_bits

VALUE = 12345678912345678989


def py_serialize_bits(value, bit_offset, bit_size, little_endian):
    bits = bit_offset + bit_size
    size = (bits + 7) // 8
    if little_endian:
        tmp = value << bit_offset
    else:
        tmp = value << -bits % 8
    # Buffer with unused bits set to zero.
    buf0 = tmp.to_bytes(size, "little" if little_endian else "big")

    # Buffer with unused bits set to one.
    buf1 = bytearray(buf0)
    if little_endian:
        # bit_offset least significant bits.
        buf1[0] |= (1 << bit_offset) - 1
        # 8 - (bit_offset + bit_size) % 8 most significant bits.
        buf1[-1] |= (0xFF00 >> -bits % 8) & 0xFF
    else:
        # bit_offset most significant bits.
        buf1[0] |= (0xFF00 >> bit_offset) & 0xFF
        # 8 - (bit_offset + bit_size) % 8 least significant bits.
        buf1[-1] |= (1 << -bits % 8) - 1

    return buf0, buf1


class TestSerialize(TestCase):
    def test_deserialize(self):
        for bit_size in range(1, 65):
            expected = VALUE & ((1 << bit_size) - 1)
            for bit_offset in range(8):
                for little_endian in [True, False]:
                    for buf in py_serialize_bits(
                        expected, bit_offset, bit_size, little_endian
                    ):
                        value = deserialize_bits(
                            buf, bit_offset, bit_size, little_endian
                        )
                        self.assertEqual(value, expected)

    def test_serialize(self):
        for bit_size in range(1, 65):
            value = VALUE & ((1 << bit_size) - 1)
            for bit_offset in range(8):
                for little_endian in [True, False]:
                    expected0, expected1 = py_serialize_bits(
                        value, bit_offset, bit_size, little_endian
                    )
                    buf = bytearray(len(expected0))
                    serialize_bits(buf, bit_offset, value, bit_size, little_endian)
                    self.assertEqual(buf, expected0)

                    buf = bytearray([0xFF] * len(expected1))
                    serialize_bits(buf, bit_offset, value, bit_size, little_endian)
                    self.assertEqual(buf, expected1)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import ctypes
import enum
import errno
from fcntl import ioctl
import functools
import mmap
import os
from pathlib import Path
import pickle
import random
import re
import signal
import socket
import stat
import string
import struct
import subprocess
import sys
import time
import traceback
from typing import NamedTuple
import unittest

from _drgn_util.platform import _IO, _IOR, NORMALIZED_MACHINE_NAME, SYS
import drgn
from tests import TestCase
from util import verrevcmp

try:
    import pyroute2

    have_pyroute2 = True

    # Before Pyroute2 commit 1eb08312de30 ("iproute/linux: try to improve flags
    # when sending del messages") (in v0.6.10), Pyroute2 incorrectly passes the
    # NLM_F_CREATE and NLM_F_EXCL flags to deletion requests, which fails with
    # ENOTSUP since Linux kernel commit 545528d78855 ("net: netlink: add
    # NLM_F_BULK delete request modifier") (in v5.19). Monkey patch the put()
    # method to strip these flags.
    if verrevcmp(getattr(pyroute2, "__version__", "0"), "0.6.10") < 0:
        # RTNL_API was named IPRouteMixin before Pyroute2 0.5.2 and was in
        # pyroute2.iproute instead of pyroute2.iproute.linux.
        try:
            from pyroute2.iproute.linux import RTNL_API as _rtnl_api
        except ImportError:
            from pyroute2.iproute import IPRouteMixin as _rtnl_api

        def _fixed_put(self, msg, msg_type, msg_flags=0x1, *args, **kwargs):
            # RTM_DEL* message types have msg_type % 4 == 1.
            if msg_type >= 16 and msg_type % 4 == 1:
                msg_flags &= ~0x600  # NLM_F_CREATE | NLM_F_EXCL
            return super(_rtnl_api, self).put(msg, msg_type, msg_flags, *args, **kwargs)

        _rtnl_api.put = _fixed_put

    # Before Pyroute2 commit f0df9a49b41f ("Fix "failed to open netns" error in
    # RISCV64.") (in v0.7.10), Pyroute2 uses the wrong syscall number on
    # RISC-V. Monkey patch the correct number.
    if verrevcmp(
        getattr(pyroute2, "__version__", "0"), "0.7.10"
    ) < 0 and os.uname().machine.startswith("riscv"):
        import pyroute2.netns

        pyroute2.netns.__NR_setns = 268
except ImportError:
    have_pyroute2 = False


class LinuxKernelTestCase(TestCase):
    prog = None
    skip_reason = None

    @staticmethod
    def _load_debug_info(prog):
        paths = []
        try:
            paths.append(os.environ["DRGN_TEST_KMOD"])
        except KeyError:
            pass
        prog.load_debug_info(paths, True)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # We only want to create the Program once for all tests, so it's cached
        # as a class variable (in the base class). If we can't run these tests
        # for whatever reason, we also cache that.
        if LinuxKernelTestCase.prog is not None:
            return
        if LinuxKernelTestCase.skip_reason is None:
            try:
                run_tests = int(os.environ["DRGN_RUN_LINUX_KERNEL_TESTS"]) != 0
            except (KeyError, ValueError):
                run_tests = True
                force_run = False
            else:
                force_run = run_tests
            if run_tests:
                prog = drgn.Program()
                try:
                    prog.set_kernel()
                except PermissionError:
                    if force_run:
                        raise
                    LinuxKernelTestCase.skip_reason = (
                        "Linux kernel tests must be run as root "
                        "(run with env DRGN_RUN_LINUX_KERNEL_TESTS=1 to force)"
                    )
                except (FileNotFoundError, ValueError):
                    if force_run:
                        raise
                    LinuxKernelTestCase.skip_reason = (
                        "Linux kernel tests require /proc/kcore "
                        "(run with env DRGN_RUN_LINUX_KERNEL_TESTS=1 to force)"
                    )
                else:
                    # Load modules that are used by test cases.
                    subprocess.check_call(
                        ["modprobe", "-a", "btrfs", "configs", "loop"]
                    )
                    try:
                        cls._load_debug_info(prog)
                        LinuxKernelTestCase.prog = prog
                        return
                    except drgn.MissingDebugInfoError as e:
                        if force_run:
                            raise
                        LinuxKernelTestCase.skip_reason = str(e)
            else:
                LinuxKernelTestCase.skip_reason = "env DRGN_RUN_LINUX_KERNEL_TESTS=0"
        raise unittest.SkipTest(LinuxKernelTestCase.skip_reason)


skip_unless_have_test_kmod = unittest.skipUnless(
    "DRGN_TEST_KMOD" in os.environ, "test requires drgn_test Linux kernel module"
)

skip_unless_have_test_disk = unittest.skipUnless(
    "DRGN_TEST_DISK" in os.environ,
    "test requires disk to overwrite; set DRGN_TEST_DISK environment variable (e.g., DRGN_TEST_DISK=/dev/vda)",
)

# Please keep this in sync with docs/support_matrix.rst.
HAVE_FULL_MM_SUPPORT = NORMALIZED_MACHINE_NAME in (
    "aarch64",
    "arm",
    "ppc64",
    "s390x",
    "x86_64",
)

skip_unless_have_full_mm_support = unittest.skipUnless(
    HAVE_FULL_MM_SUPPORT,
    f"mm support is not implemented for {NORMALIZED_MACHINE_NAME}",
)


def skip_if_slob(f):
    @skip_unless_have_test_kmod
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        if self.prog["drgn_test_slob"]:
            self.skipTest("test does not support SLOB")

    return wrapper


def skip_if_highmem(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        if self.prog["max_low_pfn"] < self.prog["max_pfn"]:
            self.skipTest("high memory is not supported")

    return wrapper


# This is a false positive if CONFIG_HIGHMEM=y and CONFIG_HIGHPTE=n, but it's
# good enough for now.
skip_if_highpte = skip_if_highmem

# Please keep this in sync with docs/support_matrix.rst.
skip_unless_have_stack_tracing = unittest.skipUnless(
    NORMALIZED_MACHINE_NAME in {"aarch64", "arm", "ppc64", "s390x", "x86_64"},
    f"stack tracing is not implemented for {NORMALIZED_MACHINE_NAME}",
)


skip_unless_have_memory_hotplug = unittest.skipUnless(
    Path("/sys/bus/memory").exists(), "memory hotplug is not supported"
)


skip_unless_have_pyroute2 = unittest.skipUnless(have_pyroute2, "pyroute2 not found")


# PRNG used by the test kernel module.
def prng32(seed):
    seed = seed.encode("ascii")
    assert len(seed) == 4
    x = int.from_bytes(seed, "big")
    while True:
        x ^= (x << 13) & 0xFFFFFFFF
        x ^= x >> 17
        x ^= (x << 5) & 0xFFFFFFFF
        yield x


def wait_until(fn, *args, **kwds):
    TIMEOUT = 5
    deadline = time.monotonic() + TIMEOUT
    sleep = 1e-6
    while True:
        if fn(*args, **kwds):
            break
        now = time.monotonic()
        if now >= deadline:
            raise Exception(f"condition was not met in {TIMEOUT} seconds")
        time.sleep(min(deadline - now, sleep))
        sleep *= 2


def proc_state(pid):
    with open(f"/proc/{pid}/status", "r") as f:
        return re.search(r"State:\s*(\S)", f.read(), re.M).group(1)


# Context manager that:
# 1. Forks a process which optionally calls a function and then stops with
#    SIGSTOP.
# 2. Waits for the child process to stop and unschedule.
# 3. Returns the PID of the child process, and return value of the function if
#    provided, from __enter__().
# 4. Kills the child process in __exit__().
@contextlib.contextmanager
def fork_and_stop(fn=None, *args, **kwds):
    with contextlib.ExitStack() as exit_stack:
        if fn:
            r, w = os.pipe()
            pipe_r = exit_stack.enter_context(open(r, "rb"))
            pipe_w = exit_stack.enter_context(open(w, "wb"))
        pid = os.fork()
        try:
            if pid == 0:
                try:
                    if fn:
                        pipe_r.close()
                        ret = fn(*args, **kwds)
                        pickle.dump(ret, pipe_w)
                        pipe_w.close()
                    while True:
                        os.kill(os.getpid(), signal.SIGSTOP)
                finally:
                    traceback.print_exc()
                    sys.stderr.flush()
                    os._exit(1)

            if fn:
                pipe_w.close()
                ret = pickle.load(pipe_r)

            _, status = os.waitpid(pid, os.WUNTRACED)
            if not os.WIFSTOPPED(status):
                raise Exception("child process exited")
            # waitpid() can return as soon as the stopped flag is set on the
            # process; see wait_task_stopped() in the Linux kernel source code:
            # https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/kernel/exit.c?h=v6.17-rc5#n1313
            # However, the process may still be on the CPU for a short window;
            # see do_signal_stop():
            # https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/kernel/signal.c?h=v6.17-rc5#n2617
            # So, we need to wait for it to fully unschedule. /proc/pid/syscall
            # contains "running" unless the process is unscheduled; see
            # proc_pid_syscall():
            # https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/fs/proc/base.c?h=v6.17-rc5#n675
            # task_current_syscall():
            # https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/lib/syscall.c?h=v6.17-rc5#n69
            # and wait_task_inactive():
            # https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/kernel/sched/core.c?h=v6.17-rc5#n2257
            syscall_path = Path(f"/proc/{pid}/syscall")
            while syscall_path.read_text() == "running\n":
                os.sched_yield()

            if fn:
                yield pid, ret
            else:
                yield pid
        finally:
            os.kill(pid, signal.SIGKILL)
            os.waitpid(pid, 0)


def parse_range_list(s):
    values = set()
    s = s.strip()
    if s:
        for range_str in s.split(","):
            first, sep, last = range_str.partition("-")
            if sep:
                values.update(range(int(first), int(last) + 1))
            else:
                values.add(int(first))
    return values


def online_cpus():
    return parse_range_list(Path("/sys/devices/system/cpu/online").read_text())


def possible_cpus():
    return parse_range_list(Path("/sys/devices/system/cpu/possible").read_text())


def meminfo_field_in_pages(name: str) -> int:
    return (
        int(
            re.search(
                rf"^{name}:\s*([0-9]+)\s*kB",
                Path("/proc/meminfo").read_text(),
                flags=re.M,
            ).group(1)
        )
        * 1024
        // mmap.PAGESIZE
    )


@contextlib.contextmanager
def temp_netns():
    while True:
        try:
            name = "".join(random.choice(string.ascii_letters) for _ in range(16))
            ns = pyroute2.NetNS(name, flags=os.O_CREAT | os.O_EXCL)
            break
        except FileExistsError:
            pass
    try:
        with ns:
            yield name, ns
    finally:
        ns.remove()


_c = ctypes.CDLL(None, use_errno=True)

_mount = _c.mount
_mount.restype = ctypes.c_int
_mount.argtypes = [
    ctypes.c_char_p,
    ctypes.c_char_p,
    ctypes.c_char_p,
    ctypes.c_ulong,
    ctypes.c_void_p,
]
MS_RDONLY = 1
MS_NOSUID = 2
MS_NODEV = 4
MS_NOEXEC = 8
MS_SYNCHRONOUS = 16
MS_REMOUNT = 32
MS_MANDLOCK = 64
MS_DIRSYNC = 128
MS_NOSYMFOLLOW = 256
MS_NOATIME = 1024
MS_NODIRATIME = 2048
MS_BIND = 4096
MS_MOVE = 8192
MS_REC = 16384
MS_SILENT = 32768
MS_POSIXACL = 1 << 16
MS_UNBINDABLE = 1 << 17
MS_PRIVATE = 1 << 18
MS_SLAVE = 1 << 19
MS_SHARED = 1 << 20
MS_RELATIME = 1 << 21
MS_KERNMOUNT = 1 << 22
MS_I_VERSION = 1 << 23
MS_STRICTATIME = 1 << 24
MS_LAZYTIME = 1 << 25


def _check_ctypes_syscall(ret, *args):
    if ret == -1:
        errno = ctypes.get_errno()
        raise OSError(errno, os.strerror(errno), *args)
    return ret


def mount(source, target, fstype, flags=0, data=None):
    _check_ctypes_syscall(
        _mount(
            os.fsencode(source),
            os.fsencode(target),
            fstype.encode(),
            flags,
            None if data is None else data.encode(),
        ),
        source,
        None,
        target,
    )


_umount2 = _c.umount2
_umount2.restype = ctypes.c_int
_umount2.argtypes = [ctypes.c_char_p, ctypes.c_int]


def umount(target, flags=0):
    _check_ctypes_syscall(_umount2(os.fsencode(target), flags), target)


_MOUNTS_RE = re.compile(
    rb"(?P<source>[^ ]+) (?P<mount_point>[^ ]+) (?P<fstype>[^ ]+) "
    rb"(?P<mount_options>[^ ]+) [0-9]+ [0-9]+"
)


class Mount(NamedTuple):
    source: str
    mount_point: Path
    fstype: str
    mount_options: str


def iter_mounts(pid="self"):
    with open(f"/proc/{pid}/mounts", "rb") as f:
        for line in f:
            match = _MOUNTS_RE.match(line)
            assert match
            yield Mount(
                source=match["source"].decode("unicode-escape"),
                mount_point=Path(match["mount_point"].decode("unicode-escape")),
                fstype=match["fstype"].decode("unicode-escape"),
                mount_options=match["mount_options"].decode("unicode-escape"),
            )


_mlock = _c.mlock
_mlock.restype = ctypes.c_int
_mlock.argtypes = [ctypes.c_void_p, ctypes.c_size_t]


_MAPS_RE = re.compile(
    rb"(?P<start>[0-9a-f]+)-(?P<end>[0-9a-f]+) (?P<flags>\S+) (?P<offset>[0-9a-f]+) (?P<dev_major>[0-9a-f]+):(?P<dev_minor>[0-9a-f]+) (?P<ino>[0-9]+)\s*(?P<path>.*)"
)


class VmMap(NamedTuple):
    start: int
    end: int
    read: bool
    write: bool
    execute: bool
    shared: bool
    offset: int
    dev: int
    ino: int
    path: str

    def is_gate(self) -> bool:
        # Arm has a gate VMA named "vectors", and x86 has one named "vsyscall".
        return self.path in ("[vectors]", "[vsyscall]")


def iter_maps(pid="self"):
    with open(f"/proc/{pid}/maps", "rb") as f:
        for line in f:
            match = _MAPS_RE.match(line)
            flags = match["flags"]
            yield VmMap(
                start=int(match["start"], 16),
                end=int(match["end"], 16),
                read=b"r" in flags,
                write=b"w" in flags,
                execute=b"x" in flags,
                shared=b"s" in flags,
                offset=int(match["offset"], 16),
                dev=os.makedev(
                    int(match["dev_major"], 16), int(match["dev_minor"], 16)
                ),
                ino=int(match["ino"]),
                path=os.fsdecode(match["path"]),
            )


def mlock(addr, len):
    _check_ctypes_syscall(_mlock(addr, len))


_prctl = _c.prctl
_prctl.argtypes = [ctypes.c_int]
_prctl.restype = ctypes.c_int


def prctl_set_vma_anon_name(addr, size, name):
    _check_ctypes_syscall(
        _prctl(
            ctypes.c_int(0x53564D41),  # PR_SET_VMA
            ctypes.c_long(0),  # PR_SET_VMA_ANON_NAME,
            ctypes.c_ulong(addr),
            ctypes.c_ulong(size),
            ctypes.c_char_p(None if name is None else os.fsencode(name)),
        )
    )


CSIGNAL = 0x000000FF
CLONE_VM = 0x00000100
CLONE_FS = 0x00000200
CLONE_FILES = 0x00000400
CLONE_SIGHAND = 0x00000800
CLONE_PIDFD = 0x00001000
CLONE_PTRACE = 0x00002000
CLONE_VFORK = 0x00004000
CLONE_PARENT = 0x00008000
CLONE_THREAD = 0x00010000
CLONE_NEWNS = 0x00020000
CLONE_SYSVSEM = 0x00040000
CLONE_SETTLS = 0x00080000
CLONE_PARENT_SETTID = 0x00100000
CLONE_CHILD_CLEARTID = 0x00200000
CLONE_DETACHED = 0x00400000
CLONE_UNTRACED = 0x00800000
CLONE_CHILD_SETTID = 0x01000000
CLONE_NEWCGROUP = 0x02000000
CLONE_NEWUTS = 0x04000000
CLONE_NEWIPC = 0x08000000
CLONE_NEWUSER = 0x10000000
CLONE_NEWPID = 0x20000000
CLONE_NEWNET = 0x40000000
CLONE_IO = 0x80000000
CLONE_CLEAR_SIGHAND = 0x100000000
CLONE_INTO_CGROUP = 0x200000000
CLONE_NEWTIME = 0x00000080


_unshare = _c.unshare
_unshare.argtypes = [ctypes.c_int]
_unshare.restype = ctypes.c_int


def unshare(flags):
    _check_ctypes_syscall(_unshare(flags))


_LOOP_SET_FD = 0x4C00
_LOOP_SET_STATUS64 = 0x4C04
_LOOP_GET_STATUS64 = 0x4C05
_LOOP_CONFIGURE = 0x4C0A
_LOOP_CTL_GET_FREE = 0x4C82

_LO_FLAGS_AUTOCLEAR = 4


def losetup(fd):
    have_loop_configure = True
    with open("/dev/loop-control", "r") as loop_control:
        while True:
            index = ioctl(loop_control.fileno(), _LOOP_CTL_GET_FREE)
            loop = open(f"/dev/loop{index}", "rb")
            close_loop = True
            try:
                # Since Linux kernel commit 3448914e8cc5 ("loop: Add
                # LOOP_CONFIGURE ioctl") (in v5.8), we can set the file
                # descriptor and the autoclear flag atomically with the
                # LOOP_CONFIGURE ioctl. Before that, we have to use
                # LOOP_SET_FD, then LOOP_GET_STATUS64 and LOOP_SET_STATUS64.
                config = bytearray(304)  # sizeof(struct loop_config)
                config[:4] = fd.to_bytes(4, sys.byteorder)
                config[60:64] = _LO_FLAGS_AUTOCLEAR.to_bytes(4, sys.byteorder)
                try:
                    if have_loop_configure:
                        ioctl(loop.fileno(), _LOOP_CONFIGURE, config)
                    else:
                        ioctl(loop.fileno(), _LOOP_SET_FD, fd)
                except OSError as e:
                    if e.errno == errno.EBUSY:
                        continue
                    elif have_loop_configure and e.errno == errno.EINVAL:
                        have_loop_configure = False
                        continue
                    raise
                if not have_loop_configure:
                    info = bytearray(232)  # sizeof(struct loop_info64)
                    ioctl(loop.fileno(), _LOOP_GET_STATUS64, info)
                    lo_flags = int.from_bytes(info[52:56], sys.byteorder)
                    lo_flags |= _LO_FLAGS_AUTOCLEAR
                    info[52:56] = lo_flags.to_bytes(4, sys.byteorder)
                    ioctl(loop.fileno(), _LOOP_SET_STATUS64, info, False)
                close_loop = False
                return loop
            finally:
                if close_loop:
                    loop.close()


def fallocate(path, offset, len):
    fd = os.open(path, os.O_WRONLY | os.O_CREAT, 0o666)
    try:
        os.posix_fallocate(fd, offset, len)
    finally:
        os.close(fd)


_swapon = _c.swapon
_swapon.argtypes = [ctypes.c_char_p, ctypes.c_int]
_swapon.restype = ctypes.c_int

_swapoff = _c.swapoff
_swapoff.argtypes = [ctypes.c_char_p]
_swapoff.restype = ctypes.c_int


def swapon(path, flags=0):
    _check_ctypes_syscall(_swapon(os.fsencode(path), flags), path)


def swapoff(path):
    _check_ctypes_syscall(_swapoff(os.fsencode(path)), path)


_BLKRRPART = _IO(0x12, 95)
_BLKSSZGET = _IO(0x12, 104)
_BLKGETSIZE64 = _IOR(0x12, 114, ctypes.sizeof(ctypes.c_size_t))


def mkswap(path, size=None):
    fd = os.open(path, os.O_WRONLY)
    try:
        if size is None:
            st = os.stat(fd)
            if stat.S_ISBLK(st.st_mode):
                size = ctypes.c_uint64()
                ioctl(fd, _BLKGETSIZE64, size)
                size = size.value
            else:
                size = st.st_size

        header = bytearray(mmap.PAGESIZE)
        header[1024:1028] = (1).to_bytes(4, sys.byteorder)  # version
        header[1028:1032] = (size // mmap.PAGESIZE - 1).to_bytes(
            4, sys.byteorder
        )  # last_page
        header[1036:1052] = os.urandom(16)  # sws_uuid
        magic = b"SWAPSPACE2"
        header[-len(magic) :] = magic

        n = os.write(fd, header)
        while n < len(header):
            n += os.write(fd, header[n:])
    finally:
        os.close(fd)


class MbrPartitionType(enum.IntEnum):
    LINUX_SWAP = 0x82
    LINUX = 0x83


class MbrPartition(NamedTuple):
    type: int
    start: int
    size: int
    bootable: bool = False


_MBR_PARTITION_STRUCT = struct.Struct("<8BII")


def write_mbr(path, partitions, sector_size=None):
    fd = os.open(path, os.O_WRONLY)
    try:
        if sector_size is None:
            sector_size = ctypes.c_int()
            ioctl(fd, _BLKSSZGET, sector_size)
            sector_size = sector_size.value

        buf = bytearray(sector_size)

        buf[0x1B8:0x1BC] = os.urandom(4)  # Disk ID.

        offset = 0x01BE
        for partition in partitions:
            if partition.start % sector_size != 0:
                raise ValueError("partition start is not sector-aligned")
            if partition.size % sector_size != 0:
                raise ValueError("partition size is not sector-aligned")
            _MBR_PARTITION_STRUCT.pack_into(
                buf,
                offset,
                0x80 if partition.bootable else 0x0,
                # Placeholder first CHS.
                0xFE,
                0xFF,
                0xFF,
                partition.type,
                # Placeholder last CHS.
                0xFE,
                0xFF,
                0xFF,
                partition.start // sector_size,
                partition.size // sector_size,
            )
            offset += _MBR_PARTITION_STRUCT.size

        buf[0x1FE] = 0x55
        buf[0x1FF] = 0xAA

        n = os.write(fd, buf)
        while n < len(buf):
            n += os.write(fd, buf[n:])
        ioctl(fd, _BLKRRPART)
    finally:
        os.close(fd)


class _perf_event_attr_sample_period_or_freq(ctypes.Union):
    _fields_ = (
        ("sample_period", ctypes.c_uint64),
        ("sample_freq", ctypes.c_uint64),
    )


class _perf_event_attr_wakeup_events_or_watermark(ctypes.Union):
    _fields_ = (
        ("wakeup_events", ctypes.c_uint32),
        ("wakeup_watermark", ctypes.c_uint32),
    )


class _perf_event_attr_config1(ctypes.Union):
    _fields_ = (
        ("bp_addr", ctypes.c_uint64),
        ("kprobe_func", ctypes.c_uint64),
        ("uprobe_path", ctypes.c_uint64),
        ("config1", ctypes.c_uint64),
    )


class _perf_event_attr_config2(ctypes.Union):
    _fields_ = (
        ("bp_len", ctypes.c_uint64),
        ("kprobe_addr", ctypes.c_uint64),
        ("probe_offset", ctypes.c_uint64),
        ("config2", ctypes.c_uint64),
    )


class perf_event_attr(ctypes.Structure):
    _fields_ = (
        ("type", ctypes.c_uint32),
        ("size", ctypes.c_uint32),
        ("config", ctypes.c_uint64),
        ("_sample_period_or_freq", _perf_event_attr_sample_period_or_freq),
        ("sample_type", ctypes.c_uint64),
        ("read_format", ctypes.c_uint64),
        ("_bitfields1", ctypes.c_uint64),
        ("_wakeup_events_or_watermark", _perf_event_attr_wakeup_events_or_watermark),
        ("bp_type", ctypes.c_uint32),
        ("_config1", _perf_event_attr_config1),
        ("_config2", _perf_event_attr_config2),
        ("branch_sample_type", ctypes.c_uint64),
        ("sample_regs_user", ctypes.c_uint64),
        ("sample_stack_user", ctypes.c_uint32),
        ("clockid", ctypes.c_int32),
        ("sample_regs_intr", ctypes.c_uint64),
        ("aux_watermark", ctypes.c_uint32),
        ("sample_max_stack", ctypes.c_uint16),
        ("__reserved2", ctypes.c_uint16),
        ("aux_sample_size", ctypes.c_uint32),
        ("__reserved3", ctypes.c_uint32),
        ("sig_data", ctypes.c_uint64),
        ("config3", ctypes.c_uint64),
    )
    _anonymous_ = (
        "_sample_period_or_freq",
        "_wakeup_events_or_watermark",
        "_config1",
        "_config2",
    )


PERF_FLAG_FD_NO_GROUP = 1 << 0
PERF_FLAG_FD_OUTPUT = 1 << 1
PERF_FLAG_PID_CGROUP = 1 << 2
PERF_FLAG_FD_CLOEXEC = 1 << 3


def perf_event_open(attr, pid, cpu, group_fd=-1, flags=PERF_FLAG_FD_CLOEXEC):
    attr.size = ctypes.sizeof(perf_event_attr)
    return _check_ctypes_syscall(
        _syscall(
            SYS["perf_event_open"],
            ctypes.byref(attr),
            ctypes.c_int(pid),
            ctypes.c_int(cpu),
            ctypes.c_int(group_fd),
            ctypes.c_ulong(flags),
        )
    )


_syscall = _c.syscall
_syscall.restype = ctypes.c_long


def create_socket(*args, **kwds):
    try:
        return socket.socket(*args, **kwds)
    except OSError as e:
        if e.errno in (errno.ENOSYS, errno.EAFNOSUPPORT, errno.ESOCKTNOSUPPORT):
            raise unittest.SkipTest("kernel does not support TCP")
        else:
            raise


IPC_PRIVATE = 0
IPC_CREAT = 0o1000
IPC_EXCL = 0o2000
IPC_NOWAIT = 0o4000

IPC_RMID = 0
IPC_SET = 1
IPC_STAT = 2
IPC_INFO = 3

_key_t = ctypes.c_int

_ftok = _c.ftok
_ftok.argtypes = [ctypes.c_char_p, ctypes.c_int]
_ftok.restype = _key_t


def ftok(path, proj_id: int) -> int:
    return _check_ctypes_syscall(_ftok(os.fsencode(path), proj_id))


_msgget = _c.msgget
_msgget.argtypes = [_key_t, ctypes.c_int]
_msgget.restype = ctypes.c_int


def msgget(key: int, msgflg: int = IPC_CREAT | 0o666) -> int:
    return _check_ctypes_syscall(_msgget(key, msgflg))


_msgctl = _c.msgctl
_msgctl.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_void_p]
_msgctl.restype = ctypes.c_int


def msgctl(msqid: int, op: int) -> None:
    _check_ctypes_syscall(_msgctl(msqid, op, 0))


_semget = _c.semget
_semget.argtypes = [_key_t, ctypes.c_int, ctypes.c_int]
_semget.restype = ctypes.c_int


def semget(key: int, nsems: int, semflg: int = IPC_CREAT | 0o666) -> int:
    return _check_ctypes_syscall(_semget(key, nsems, semflg))


_semctl = _c.semctl
_semctl.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int]
_semctl.restype = ctypes.c_int


def semctl(semid: int, semnum: int, op: int) -> None:
    _check_ctypes_syscall(_semctl(semid, semnum, op))


SHM_LOCK = 11


_shmget = _c.shmget
_shmget.argtypes = [_key_t, ctypes.c_int, ctypes.c_int]
_shmget.restype = ctypes.c_int


def shmget(key: int, size: int, shmflg: int = IPC_CREAT | 0o666) -> int:
    return _check_ctypes_syscall(_shmget(key, size, shmflg))


_shmctl = _c.shmctl
_shmctl.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_void_p]
_shmctl.restype = ctypes.c_int


def shmctl(shmid: int, op: int) -> None:
    _check_ctypes_syscall(_shmctl(shmid, op, 0))


_aio_context_t = ctypes.c_ulong


def io_setup(nr_events):
    ctx_id = _aio_context_t()
    _check_ctypes_syscall(
        _syscall(SYS["io_setup"], ctypes.c_uint(nr_events), ctypes.byref(ctx_id))
    )
    return ctx_id.value


def io_destroy(ctx_id):
    _check_ctypes_syscall(_syscall(SYS["io_destroy"], _aio_context_t(ctx_id)))


_aio_key_and_aio_rw_flags = [
    ("aio_key", ctypes.c_uint32),
    ("aio_rw_flags", ctypes.c_int),
]
if sys.byteorder == "big":
    _aio_key_and_aio_rw_flags.reverse()


IOCB_CMD_PREAD = 0
IOCB_CMD_PWRITE = 1
IOCB_CMD_FSYNC = 2
IOCB_CMD_FDSYNC = 3
IOCB_CMD_POLL = 5
IOCB_CMD_NOOP = 6
IOCB_CMD_PREADV = 7
IOCB_CMD_PWRITEV = 8


class iocb(ctypes.Structure):
    _fields_ = (
        ("aio_data", ctypes.c_uint64),
        *_aio_key_and_aio_rw_flags,
        ("aio_lio_opcode", ctypes.c_uint16),
        ("aio_reqprio", ctypes.c_int16),
        ("aio_fildes", ctypes.c_uint32),
        ("aio_buf", ctypes.c_uint64),
        ("aio_nbytes", ctypes.c_uint64),
        ("aio_offset", ctypes.c_int64),
        ("aio_reserved2", ctypes.c_uint64),
        ("aio_flags", ctypes.c_uint32),
        ("aio_resfd", ctypes.c_uint32),
    )


def io_submit(ctx_id, iocbs):
    nr = len(iocbs)
    iocbp = (ctypes.POINTER(iocb) * nr)(*(ctypes.pointer(iocb) for iocb in iocbs))
    return _check_ctypes_syscall(
        _syscall(
            SYS["io_submit"],
            _aio_context_t(ctx_id),
            ctypes.c_long(nr),
            ctypes.byref(iocbp),
        )
    )
